#Newer FlashWriter SPL thingie
import io
import time
from parse import parse
from .base import PartitionBase, FlashDeviceBase

class WriterFLW():
    devices = {}
    version = 0
    next_buffer = 0

    def __init__(self, terminal):
        self.terminal = terminal
        r = self.terminal.wait("FlashWriter ({}) (help for useless info)", timeout=100)
        self.version = r[0]

    def edcl_send_data(self, fd, length, cb):
        self.next_buffer = 0
        total = length
        while length:
            self.edcl_sync_wait(0)
            sent = self.edcl_submit_buffer(fd, length)
            length -= sent
            if cb:
                cb(total, total - length, sent)

    def edcl_read_data(self, fd, length, cb):
        self.next_buffer = 0
        self.edcl_sync_wait(0)
        curbuf = self.edcl_get_buffer()
        self.edcl_sync(curbuf)
        total = length
        while length > 0:
            last = not length > self.buffer_length
            data = self.edcl_fetch_buffer(last)
            if len(data) > length:
                data = data[:length]
            fd.write(data)
            length -= len(data)
            if cb:
                cb(total, total - length, len(data))


    def edcl_get_buffer(self):
        return self.addr_buffers[self.next_buffer]

    def edcl_next(self):
        self.next_buffer += 1
        self.next_buffer &= 1

    buffers_count = 0

    def edcl_fetch_buffer(self, is_last):
        self.edcl_sync_wait(0)
        curbuf = self.edcl_get_buffer()
        #print(f"Read buffer {curbuf:x} cnt: {self.buffers_count}")
        self.edcl_next()
        data = self.terminal.xfer.read(curbuf, self.buffer_length)
        if not is_last:
            self.edcl_sync(self.edcl_get_buffer())
        return data

    def edcl_submit_buffer(self, fd, maxlength):
        index = self.next_buffer & 1
        toread = self.buffer_length
        if maxlength < self.buffer_length:
            toread = maxlength 

        data = fd.read(toread)
        #print(f"Filling buffer {index} with {len(data)} bytes addr {self.addr_buffers[index]:x} pos {fd.tell()} cnt: {self.buffers_count}")
        if len(data) > 0:
            self.terminal.xfer.write(self.addr_buffers[index], data)
            self.buffers_count += 1
            self.edcl_sync(self.edcl_get_buffer())
            self.edcl_next()
            return toread
        raise Exception("Out of bounds reading input stream")

    def swap32(self, x):
        return int.from_bytes(x.to_bytes(4, byteorder='little'), byteorder='big', signed=False)

    def dumpuart(self):
        # Debug only
        #if self.terminal.ser.inWaiting():
        #    print(self.terminal.ser.readline())
        pass

    oops = 0
    def edcl_sync(self, value=None):
        if value == None:
            ret = self.terminal.xfer.read32(self.addr_sync)
            ret = self.swap32(ret)
        else:
            while True: 
                self.terminal.xfer.write32(self.addr_sync, self.swap32(value))
                tmp = self.terminal.xfer.read32(self.addr_sync_ack)
                tmp = self.swap32(tmp)
                if tmp == value:
                    self.terminal.xfer.write32(self.addr_sync_ack, 0)                    
                    break
            ret = value
        return ret

    def edcl_sync_wait(self, val, tout=5):
        value = 0
        start = time.monotonic()
        while time.monotonic() - start < 15:
            value = self.edcl_sync()
            if  value == val:
                return
        raise Exception(f"EDCL Sync timeout, last value {value:x}, want {val:x}")

    def __getitem__(self, key):
        self.select(key)
        return self.devices[key]
    
    def select(self, device):
        return self.terminal.cmd(f"select {device}", "Device {} selected")

    def discover(self):
        devices = {}
        ret,lines = self.terminal.cmd("bufptr", "buffers {:x} {:x} sync {:x} ack {:x} length {:x}")
        self.addr_buffers = [ ret[0], ret[1] ]
        self.addr_sync = ret[2]
        self.addr_sync_ack = ret[3]
        self.buffer_length = ret[4]

        ret, lines = self.terminal.cmd("list", "completed")
        for l in lines:
            ret = parse("id: {} part: {} size: {:x} erasesize: {:x} writesize: {:x}", l)
            if ret:
                devices[ret[0]] = {                
                    "part" :        ret[1],
                    "size" :        ret[2],
                    "write_size" :  ret[3],
                    "read_size" :   ret[4],
                    "erase_size" :  ret[3]
                }
        return devices

class FlashDeviceFLW(FlashDeviceBase):
    writer = None
    skip_erase_before_write = True
    def __init__(self, terminal, device):
        self.name = device["device"]
        super().__init__()
        self.writer = WriterFLW(terminal)
        devices = self.writer.discover()
        if not self.name in devices:
            raise Exception(f"Failed to find device {self.name} in devices list")
        for key, value in devices[self.name].items():
            setattr(self, key, value)
        self.writer.select(self.name)

    def switchbaud(self, newbaud):
        self.terminal.cmd(f"baudrate {newbaud:d}", "Baudrate: current={:d},new={:d}")
        self.terminal.reopen(speed=newbaud)

    def _read_edcl(self, fd, offset, length, cb):
        self.terminal.cmd(f"duplicate E {offset:x} {length:x}", "ready", timeout=120)
        return self.writer.edcl_read_data(fd, length, cb)

    def _read_xmodem(self, fd, offset, length, cb):
        self.terminal.cmd(f"duplicate X {offset:x} {length:x}", "ready", timeout=120)
        return self.terminal.xfer.to_file(0, length, fd, callback = cb)

    def _read(self, fd, offset, length, cb):
        if self.terminal.xfer.how == "xmodem":
            return self._read_xmodem(fd, offset, length, cb)
        elif self.terminal.xfer.how == "edcl":
            return self._read_edcl(fd, offset, length, cb) 

    def _write_xmodem(self, fd, offset, length, cb):
        self.terminal.cmd(f"program X {offset:x} {length:x}", "completed", timeout=120)
        self.terminal.xfer.from_file(0, fd, cb, -1, length)

    def _write_edcl(self, fd, offset, length, cb):
        self.terminal.cmd(f"program E {offset:x} {length:x}", "completed", timeout=120)
        self.writer.edcl_send_data(fd, length, cb)

    def _write(self, fd, offset, length, callback = None):
        if self.terminal.xfer.how == "xmodem":
            return self._write_xmodem(fd, offset, length, callback)
        elif self.terminal.xfer.how == "edcl":
            return self._write_edcl(fd, offset, length, callback) 

    def _erase(self, offset, length, callback = None):
        self.terminal.cmd(f"erase {offset:x} {length:x}", "completed")

    def __str__(self):
        return f'FlashWriter Device {self.name} part: {self.part} size {self.size} erase_size: {self.erase_size} write_size: {self.write_size}'


#TODO: MBR Partitions
class FlasherFlwMMC(PartitionBase, FlashDeviceFLW):
    device = "mmc{}"
    protocol = "flashwriter"
    def __init__(self, terminal, device):
        PartitionBase.__init__(self, terminal)
        FlashDeviceFLW.__init__(self, terminal, device)

#TODO: DTB Partitions
class FlasherFlwSF(PartitionBase, FlashDeviceFLW):
    device = "sf{}"
    protocol = "flashwriter"
    def __init__(self, terminal, device):
        PartitionBase.__init__(self, terminal)
        FlashDeviceFLW.__init__(self, terminal, device)

class FlasherFlwNOR(PartitionBase, FlashDeviceFLW):
    device = "nor{}"
    protocol = "flashwriter"
    def __init__(self, terminal, device):
        PartitionBase.__init__(self, terminal)
        FlashDeviceFLW.__init__(self, terminal, device)

class FlasherFlwI2C(PartitionBase, FlashDeviceFLW):
    device = "i2c{:d}-{:x}"
    protocol = "flashwriter"
    def __init__(self, terminal, device):
        PartitionBase.__init__(self, terminal)
        FlashDeviceFLW.__init__(self, terminal, device)

class FlasherFlwNAND(PartitionBase, FlashDeviceFLW):
    device = "nand{}"
    protocol = "flashwriter"
    def __init__(self, terminal, device):
        PartitionBase.__init__(self, terminal)
        FlashDeviceFLW.__init__(self, terminal, device)



from .base import PartitionBase, FlashDeviceBase
class FlashDeviceDumb(FlashDeviceBase):
    def __init__(self):
        self.erase_size = 1
        self.write_size = 1
        self.read_size = 1
        self.size = 128 * 1024 * 1024
        self.part = "dumb flash device"
        super().__init__()

    def _read(self, fd, offset, length, callback = None):
        raise Exception("DUMB SPLs DO NOT SUPPORT READBACK")

    def _write(self, fd, offset, length, callback = None):
        if offset != 0:
            raise Exception(f"DUMB SPLs DO NOT SUPPORT OFFSET != 0 (current is {offset}")
        self.terminal.wait("boot: Press 'X' and send me the image")
        self.terminal.write("X".encode("ascii"))
        if self.terminal.xfer.how != "xmodem" and self.terminal.xfer.how != "xmodem1k":
            raise Exception("Dumb spls only support xmodem")
        return self.terminal.xfer.from_file(0, fd, callback, offset, length)

    def _erase(self, offset=0, length=-1, callback = None):
        pass

    def switchbaud(self, newbaud):
        raise Exception("Not implemented")

class FlasherDumb(PartitionBase, FlashDeviceDumb):
    device = "dumb"
    protocol = "dumb"
    def __init__(self, terminal, device):
        FlashDeviceDumb.__init__(self)
        PartitionBase.__init__(self, terminal)



from rumboot.testing.context import TestContext
from rumboot.testing.registry import test_registry
from rumboot.testing.testing_cli import start_testing_cli
from rumboot.testing.testing_gui import start_testing_gui


# starts before test loading
_test_context = TestContext()
_test_context.process_cmdline()
_test_context.load_environment()


# starts after test loading
def rumboot_start_testing():
    _test_context.setup_environment()
    if _test_context.env["gui"]:
        start_testing_gui(test_registry, _test_context)
    else:
        start_testing_cli(test_registry, _test_context)




# _states are:
#    POWER
#    RESET
#    HOST
#    ...

import time
class base:
    name = "None"
    silent = False
    resetdelay = 1
    warned = False
    supported = ["POWER", "RESET"]
    chip = None

    def __init__(self, terminal, opts):
        if opts["port"].find("socket") != -1:
            self.silent = True
        self._states = {}
        self.terminal = terminal

    def __getitem__(self, key):
        if not key in self.supported:
            raise Exception(f"Control {key} in unknown")
        return self._states[key]

    def __setitem__(self, key, value):
        if not key in self.supported:
            raise Exception(f"Control {key} in unknown")
        self._states[key] = value
        
        if self.name == "None" and not self.silent and not self.warned:
            print("Please, power-cycle board...")
            self.warned = True #Show it only once

    def set_chip(self, chip):
        self.chip = chip

    def get_options(self):
        return {}

    def add_argparse_options(self, parser):
        for key,opts in self.get_options(self).items():
            parser.add_argument("--" + key,
                    **opts)

# Legacy API wrappers
    def power(self, on):
        self["POWER"] = on

# Power and reset
# POWER: 1 - active, 0 - offline
# RESET: 1 - active, device is in reset, 0 - normal operation 
    def reset(self):
        self["POWER"] = 0 #Disable power
        self["RESET"] = 1 #Assert reset
        time.sleep(1)
        self["POWER"] = 1 #Enable power
        time.sleep(0.2)
        self["RESET"] = 0 #Remove reset
 
    def resetToHost(self, flags = []):
        self["HOST"] = 1
        self.reset()

    def resetToNormal(self, flags = []):
        self["HOST"] = 0
        self.reset()



import os
import time
import usb.core
import pygpiotools
import usb.util
from parse import parse
from rumboot.resetseq.resetSeqBase import base

class pl2303(base):
    name = "PL2303HX"
    swap   = False
    supported = ["POWER", "RESET"]
    mapping = {
        "POWER" : 0,
        "RESET" : 1
    }

    def __init__(self, terminal, opts):
        self.invert_power   = opts["pl2303_invert_power"]
        self.invert_reset   = opts["pl2303_invert_reset"]
        self.swap           = opts["pl2303_swap"]
        self.__handle = pygpiotools.connect_pyserial("pl2303", terminal.ser)

        if self.swap:
            tmp = self.mapping["POWER"]
            self.mapping["POWER"] = self.mapping["RESET"]
            self.mapping["RESET"] = tmp

        super().__init__(terminal, opts)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

        # We have power inverted on all boards compared to base
        # so let's make sure it works with default options
        if not self.invert_power and key == "POWER":
            value = not value
        if self.invert_reset and key == "RESET":
            value = not value

        pygpiotools.direction(self.__handle, self.mapping[key], "OUTPUT")
        pygpiotools.write(self.__handle, self.mapping[key], value)

    def get_options(self):
        return {
                "pl2303-invert-reset" : {
                    "help" : "Invert pl2303 reset signal",
                    "default" : False,
                    "action"  : 'store_true'
                },
                "pl2303-invert-power" : {
                    "help" : "Invert pl2303 power signal",
                    "default" : False,
                    "action"  : 'store_true'
                },
                "pl2303-swap" : {
                    "help" : "Swap pl2303 reset and power mapping",
                    "default" : False,
                    "action"  : 'store_true'
                }
            }


#NAND on MB7705. Legacy 10+ year stuff, but we're all about LTS

import time
import io
import hexdump
import humanfriendly
from parse import parse
from .base import PartitionBase, FlashDeviceBase

class PartitionMboot(PartitionBase):
    partitions_synced = False

    def __init__(self, terminal, name, size=None):
        super().__init__(terminal)
        self.name = name

        #Let's register internal partitions
        #This logic as the same, as in mboot
        bootsz = 256 * 1024
        if self.erase_size > bootsz:
            bootsz = self.erase_size

        self.add_partition("boot", 0, bootsz)
        self.add_partition("env",  bootsz, self.erase_size)
        self.add_partition("dtb",  bootsz + self.erase_size, self.erase_size)
        self.add_partition("firmware", 0, self.size)

    def get_real_offset(self):
        return self.__offset

    def commit(self, save=False):
        if self.partitions_synced:
            return

        parts=""
        sizes = []
        for k,v in self.partitions.items():
            if k == "boot":
                continue
            if k == "env":
                continue
            if k == "dtb":
                continue
            if k == "firmware":
                continue
            parts = f"{parts}{k},"
            sizes.append(f"setenv {k}_size {v.size}")

        self.mboot_cmd(f"setenv parts {parts}")
        for s in sizes:
            self.mboot_cmd(s)
        self.mboot_cmd("partscan")
        if save:
            self.mboot_cmd("saveenv")

class FlashDeviceMB7707(FlashDeviceBase):
    cmdaddr    = 0
    magic_addr = 0x00100000
    mbootoffs  = 0x00100010
    TEXT_BASE  = 0x00100014
    RUNPTR     = 0x00100010
    RAM_BASE   = 0x40100000


    def extract_nand_info(self):
        fmt = "mnand_read_id: CS{:d} NAND {} {} {} size({:d}) writesize({:d}) oobsize({:d}) erasesize({:d})"
        chips = []
        _id, lines = self.terminal.wait("mnand_read_id: flash id {:x}")

        if _id[0] > 0:
            params, lines = self.terminal.wait(fmt)
            chips.append(params)

        _id, lines = self.terminal.wait("mnand_read_id: flash id {:x}")
        if _id[0] > 0:
            params, lines = self.terminal.wait(fmt)
            chips.append(params)
        
        size = 0
        part = ""
        for c in chips:
            size += humanfriendly.parse_size(c[1])

        if len(chips) > 1:
            l = "(x2)"
        else:
            l = "(x1)"
        attrs = {
            "size" : size,
            "part" : f"NAND {chips[0][1]} {chips[0][2]} {chips[0][3]} {l}",
            "read_size" : chips[0][4],
            "write_size" : chips[0][5],
            "oob_size" : chips[0][6],
            "erase_size" : chips[0][7],
        }
        self.nand_chips = chips
        return attrs

    def __init__(self, terminal, device):
        self.name = device
        super().__init__()
        self.magic(0xdeadc0de) # Enable slave mode
        
        self.wait_nmagic(0xdeadc0de);
        self.cmdaddr = self.magic()

        layout = self.extract_nand_info()
        for key, value in layout.items():
            setattr(self, key, value)

    def magic(self, value = None):
        if value != None:
            self.terminal.xfer.write32(self.magic_addr, value)
        else:
            value = -1
        ret = self.terminal.xfer.read32(self.magic_addr)
        return ret

    def wait_magic(self, value):
        while True:
            v = self.terminal.xfer.read32(self.magic_addr)
            time.sleep(0.1)
            if v == value:
                break

    def wait_nmagic(self, value):
        while True:
            v = self.terminal.xfer.read32(self.magic_addr)
            time.sleep(0.1)
            if v != value:
                break
    
    def get_buffer(self):
       self.wait_nmagic(0xa1)
       return self.magic()

    def mboot_cmd(self, cmd):
       data = cmd.encode("ascii")
       data += b"\x00"
       self.terminal.xfer.write(self.cmdaddr, data)
       self.magic(0xdeadbeaf)
       self.wait_nmagic(0xdeadbeaf)

    def dumpuart(self):
        while self.terminal.ser.inWaiting():
            print(self.terminal.ser.readline())
 
    def _write(self, fd, offset, length, callback = None):
        self.commit()
        self.mboot_cmd(f"eupgrade 0x{offset:x} 0x{length:x}")
        es = self.magic()
        self.magic(0xa1)
        total = self.stream_size(fd)
        chunksize = self.erase_size
        while length > 0:
            buffer_addr = self.get_buffer()
            data = fd.read(chunksize)
            self.terminal.xfer.write(buffer_addr, data)
            sent = len(data)
            length -= sent
            self.magic(0xa1)
            if callable(callback):
                callback(total, total - length, sent)
        self.magic(0)
        self.wait_nmagic(0)

    def get_bad_blocks(self):
        self.mboot_cmd("mtd bad; version")
        self.terminal.wait("MTD mnand bad blocks:")
        ret, lines = self.terminal.wait("mboot-{}")
        bads = []
        for l in lines:
            try:
                bads.append(int(l, 16))
            except:
                pass
        return bads

    def _read(self, fd, offset, length, callback = None):
        self.commit()
        bads = self.get_bad_blocks()
        maxchunksize = self.erase_size
        total = length
        while length > 0:
            if length > maxchunksize:
                chunk = maxchunksize
            else:
                chunk = length
            if not offset in bads:
                self.mboot_cmd(f"mtd read 0x40100000 0x{offset:x} 0x{chunk:x}")
                data = self.terminal.xfer.read(0x40100000, chunk)
                fd.write(data)
                length -= len(data)
                offset += len(data)
                if callback(callback):
                    callback(total, total - length, chunk)
            else:
                offset += self.erase_size

    def _erase(self, offset=0, length=-1, callback = None):
        if length == -1:
            length = self.size
        self.terminal.cmd(f"mtd scrub 0x{offset:x} {length:x}")

    def switchbaud(self, newbaud):
        raise Exception("Not implemented for this board")

class Flasher7707NAND(PartitionMboot, FlashDeviceMB7707):
    device = "nand"
    protocol = "legacy_edcl"
    #For MB77.07 device must equal partition, use firmware for full-flash
    def __init__(self, terminal, device):
        self.terminal = terminal #We need that early
        FlashDeviceMB7707.__init__(self, terminal, device)
        PartitionMboot.__init__(self, terminal, device)



import sys
from rumboot.testing.user_interaction import *
from rumboot.testing.core import *
from rumboot.testing.test_desc import *
from rumboot.testing.executor import TestExecutor
from rumboot.testing.reports import JUintReport


class _TestingCLI(UserInteraction):

    def __init__(self, test_registry, test_context):
        self._test_registry = test_registry
        self._test_context = test_context
        self._reports = None
        if self._test_context.env["report_file_path"]:
            self._reports = JUintReport(self._test_context.env["report_file_path"])
        self._executor = TestExecutor(test_context, self)

    # UserInteraction
    def request_message(self, test, text):
        input(f"{test.full_name} - {text} - Press <ENTER>")

    # UserInteraction
    def request_yes_no(self, test, text):
        answer = input(f"{test.full_name} - {text} - (Y/N)?")
        if answer in ["Y", "y"]:
            return True
        if answer in ["N", "n"]:
            return False
        print("Unknown answer, interrpreted as NO")
        return False

    # UserInteraction
    def request_option(self, test, text, options):
        print(f"{test.full_name} - {text}:")
        index = 1
        for opt in options:
            print(f"{index} - {opt}")
            index += 1
        answer = input()
        try:
            index = int(answer)
        except:
            index = 0
        if index < 1 or index > len(options):
            print(f"Unknown answer, interrpreted as {options[0]}")
            return 0
        return index - 1

    def execute(self):
        update_suitable(self._test_registry, self._test_context)
        self._stat_all_tests = 0
        self._stat_skipped = 0
        self._stat_fault = 0
        self._stat_passed = 0
        self._test_registry.test_iteration(self._execute_one_test)
        print(f"==========")
        print(f"All tests : {self._stat_all_tests}")
        print(f"Skipped   : {self._stat_skipped}")
        print(f"Passed    : {self._stat_passed}")
        print(f"Fault     : {self._stat_fault}")
        if self._reports:
            print("Saving reports...")
            self._reports.flush()

    def _execute_one_test(self, test_desc):
        self._stat_all_tests += 1
        self._executor.exec_test(test_desc)
        if test_desc.status == TEST_STATUS_NOT_EXECUTED:
            self._stat_skipped += 1
        elif test_desc.status == TEST_STATUS_PASSED:
            self._stat_passed += 1
        elif test_desc.status == TEST_STATUS_FAULT:
            self._stat_fault += 1
        else:
            raise Exception("Unknown test status")
        if self._reports:
            self._reports.add_test_result(test_desc)


def start_testing_cli(test_registry, test_context):
    test_cli = _TestingCLI(test_registry, test_context)
    test_cli.execute()



import sys
import io
import threading
import inspect
import ctypes
import time
import traceback
from rumboot.terminal import terminal
from rumboot.testing.test_desc import *


class TestExecutor:

    def __init__(self, test_context, user_interaction, reports=None, log_func=None):
        self._test_context = test_context
        self._user_interaction = user_interaction
        self._reports = reports
        self._log_func = log_func
        self._log_stream = None

    def exec_test(self, test_desc):
        test_desc.status == TEST_STATUS_NOT_EXECUTED
        test_desc.log_text = None

        self._log_stream = io.StringIO()
        self._save_stdout = sys.stdout
        self._save_stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self
        try:
            print(f"=== Processing {test_desc.full_name} ===")
            if test_desc.suitable:
                self._test_execution_with_log(test_desc)
            if test_desc.status == TEST_STATUS_NOT_EXECUTED:
                print("The test is not suitable for the environment")
            elif test_desc.status == TEST_STATUS_PASSED:
                print("Passed")
            elif test_desc.status == TEST_STATUS_FAULT:
                print("Fault")
            else:
                raise Exception("Unknown test status")
        finally:
            sys.stdout = self._save_stdout
            sys.stderr = self._save_stderr
            self._save_stdout = None
            self._save_stderr = None
        test_desc.log_text = self._log_stream.getvalue()
        self._log_stream = None

    # IOBase
    def write(self, text):
        self._log_stream.write(text)
        if self._log_func:
            self._log_func(text)
        else:
            print(text, file=self._save_stdout, end="")

    # IOBase
    def flush(self):
        pass

    def _test_execution_with_log(self, test_desc):
        test_desc.status = TEST_STATUS_FAULT
        timeout_sec = test_desc.test_class.timeout
        if "timeout" in test_desc.params:
            timeout_sec = test_desc.params["timeout"]

        thread = threading.Thread(target=self._test_execution_in_thread, args=[test_desc])
        thread.wait_user = False
        thread.start()
        time_left = timeout_sec
        while time_left > 0:
            thread.join(timeout=1)
            if not thread.is_alive():
                break
            if not thread.wait_user:
                time_left -= 1

        if thread.is_alive():
            while thread.is_alive():
                _async_raise(thread.ident, SystemExit)
                time.sleep(0.1)
            thread.join()
            test_desc.status = TEST_STATUS_FAULT
            print(f"The test has been terminated by timeout {timeout_sec} seconds")

    def _test_execution_in_thread(self, test_desc):
        params_for_xfers = {
            "default": self._test_context.env["connection"]["transport"],
            "force_static_arp": self._test_context.env["connection"]["force_static_arp"],
            "edcl_timeout": self._test_context.env["connection"]["edcl_timeout"]
        }
        if self._test_context.env["connection"]["edcl_ip"]:
            params_for_xfers["edcl_ip"] = self._test_context.env["connection"]["edcl_ip"]
        if self._test_context.env["connection"]["edcl_mac"]:
            params_for_xfers["edcl_mac"] = self._test_context.env["connection"]["edcl_mac"]
        term = terminal(self._test_context.env["connection"]["port"], self._test_context.env["connection"]["baud"], xferparams=params_for_xfers)
        term.set_chip(self._test_context.chip)
        term.xfer.selectTransport(self._test_context.env["connection"]["transport"])

        reset = self._test_context.resets[self._test_context.env["reset"]["name"]](term, self._test_context.env["reset"])

        try:
            test = test_desc.test_class(test_desc.name, test_desc.full_name, term, reset, self._test_context.env, test_desc.params, self._user_interaction)
            if test.run():
                test_desc.status = TEST_STATUS_PASSED
        except Exception:
            print(traceback.format_exc())
        # ??? for term deleting (must be fixed in the terminal class)
        term.xfer.xfers = None
        term.xfer.how = None
        term.xfer.xfer = None
        term.xfer = None
        term.opf.objects = None
        term.opf = None
        term.ser.write = None
        term.ser = None
        # ???


def _async_raise(tid, exctype):
    '''Raises an exception in the threads with id tid'''
    if not inspect.isclass(exctype):
        raise TypeError("Only types can be raised (not instances)")
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("invalid thread id")
    elif res != 1:
        # "if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
        raise SystemError("PyThreadState_SetAsyncExc failed")



import argparse
import os
import sys

from rumboot.chipDb import ChipDb
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.resetSeq import ResetSeqFactory
from rumboot.terminal import terminal

try:
    import yaml
except:
    yaml = None

class arghelper():
    configs = [
        "~/.rumboot.yaml", 
        "/etc/rumboot.yaml", 
        os.path.dirname(__file__) + "/../rumboot.yaml"
        ]
    config = None
    default_port = "/dev/ttyUSB0"

    def get_default_port(self, chip):
        try:
            ret = self.config['xrun']['chips'][chip.__name__]["port"]
            return ret
        except:
            pass
        
        try: 
            ret = self.config['xrun']['defaults']["port"]
            return ret
        except:
            return self.default_port

    def get_default_baud(self, chip):
        try:
            ret = self.config['xrun']['chips'][chip.name]["baudrate"]
            return ret
        except:
            if chip != None:
                return chip.baudrate
            return 115200

    def __init__(self):
        if yaml == None:
            print("[W] No yaml, disabling config files. Install with 'pip install yaml' to fix")
            return
        for conf in self.configs:
            conf = os.path.expanduser(conf)
            if (os.path.exists(conf)):
                self.load_config(conf)
                break

    def load_config(self, conf):
        print("[!] Using configuration file: " + conf)
        with open(conf, 'r') as stream:
            try:
                self.config = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)

    def process_files(self, files, rebuild):
        ret = files
        dumps = { }
        for k,f in enumerate(files):
            target = f.replace(".bin", ".all")
            dmp = f.replace(".bin", ".dmp")
            if rebuild:
                bres = os.system("cmake --build . --target %s" % target)
                if (bres != 0):
                    sys.exit(bres)

            ret[k] = open(f, "rb")

            try:
                dumps[f] = open(dmp, "r")
            except:
                print("WARN: Missing dump file: " + dmp + ". Detailed traces will be unavailable")
                pass

        return ret, dumps

    def add_file_handling_opts(self, parser, need_file=False):
        group = parser.add_argument_group('File Handling')
        parser.add_argument('-f','--file',action='append',nargs=1,
                        type=str,
                        help="Image file (may be specified multiple times)")
        group.add_argument("-c", "--chip_id",
                        help="Override chip id (by name or chip_id)",
                        nargs=1, metavar=('chip_id'),
                        required=False)            

    def add_terminal_opts(self, parser):
        group = parser.add_argument_group('Connection Settings')
        group.add_argument("-l", "--log",
                        help="Log terminal output to file",
                        type=argparse.FileType("w+"),
                        required=False)
        group.add_argument("-p", "--port",
                        help="Serial port to use",
                        metavar=('port'),
                        required=False),
        group.add_argument("-b", "--baud",
                        help="Serial line speed",
                        type=int,
                        metavar=('speed'),
                        required=False)
        group.add_argument("-e", '--edcl',
                        help="Use edcl for data uploads (when possible)",
                        action="store_true",
                        default=False,
                        required=False)
        group.add_argument('--force-static-arp',
                        help="Always add static ARP entries",
                        action="store_true",
                        default=False,
                        required=False)
        group.add_argument('--edcl-ip',
                        help="Use specific EDCL IP address (default - from chip id)",
                        default=None,
                        required=False
                        )
        group.add_argument('--edcl-mac',
                        help="Use specific EDCL MAC address (default - from chip id)",
                        default=None,
                        required=False
                        )
        group.add_argument('--edcl-timeout', type=float,
                        help="Use specific EDCL MAC address (default - from chip id)",
                        default=7.0,
                        required=False
                        )


    def add_resetseq_options(self, parser, rfactory):
        group = parser.add_argument_group('Reset Sequence options', "These options control how the target board will be reset")
        resetlist = ""
        for r in rfactory.classes:
            resetlist = resetlist + " " + r
        resetlist = resetlist.strip()
        group.add_argument("-r", "--reset",
                        help="Reset sequence to use (%s)" % resetlist,
                        nargs=1, metavar=('method'),
                        default="none",
                        required=False)
        for name, sequence in rfactory.classes.items():
            if hasattr(sequence, "add_argparse_options"):
                g = parser.add_argument_group(name + " reset sequence options")
                sequence.add_argparse_options(sequence, g)
        
    def detect_terminal_options(self, opts, chip):
        # Now, let's update default options, if needed
        if opts.port == None:
            opts.port = self.get_default_port(chip)
        if opts.baud == None:
            opts.baud = self.get_default_baud(chip)

    def detect_chip_type(self, opts, chips, formats):
        c = None
        try:
            t = formats.guess(opts.file[0][0])
            c = chips[t.get_chip_id()]
        except:
            pass

        if opts.chip_id != None:
            c = chips[opts.chip_id[0]]

        if c == None:
            print("ERROR: Failed to auto-detect chip type")
            print("HINT: Specify a file (-f) or set chip id manually (-c)")

        self.detect_terminal_options(opts, c)
        return c

    def create_core_stuff_from_options(self, opts):
            resets  = ResetSeqFactory("rumboot.resetseq")
            formats = ImageFormatDb("rumboot.images")
            chips   = ChipDb("rumboot.chips")
            c = self.detect_chip_type(opts, chips, formats)
            if not c:
                return None, None, None
            print("Detected chip:              %s (%s)" % (c.name, c.part))

            if c == None:
                raise Exception("ERROR: Failed to auto-detect chip type")
            if opts.baud == None:
                opts.baud = c.baudrate

            params_for_xfers = {
                "force_static_arp" : opts.force_static_arp,
                "default" : "xmodem"
            }

            if opts.edcl:
                params_for_xfers["default"] = "edcl"
            if opts.edcl_ip:
                params_for_xfers["edcl_ip"] = opts.edcl_ip
            if opts.edcl_mac:
                params_for_xfers["edcl_mac"] = opts.edcl_mac
            if opts.edcl_timeout:
                params_for_xfers["edcl_timeout"] = opts.edcl_timeout

            term = terminal(opts.port, opts.baud, xferparams = params_for_xfers)
            reset = resets[opts.reset[0]](term, vars(opts))
            term.set_chip(c)
            reset.set_chip(c)

            print("Reset method:               %s" % (reset.name))
            print("Baudrate:                   %d bps" % int(opts.baud))
            print("Port:                       %s" % opts.port)
            print("Preferred data transport:   %s" % params_for_xfers["default"])

            return c, term, reset



class MacGenerator():
    macs = {}

    def __init__(self, chip_id = 0, board_id = 0, oui = "EC:17:66"):
        self.oui = oui
        self.chip_id = chip_id
        self.board_id = board_id

    def get(self, id: int, iface :int):
        if not id in self.macs:
            self.macs[id] = [ chip_id % 256, board_id % 256, random.randrange(0,255) & 0xF8 ]
        b = self.macs[id]
        return f"{self.oui}:{b[0]:X}:{b[1]:X}:{b[2] | iface :X}"



__version__ = "0.9.22"
__copyright__ = """
(C) 2018-2021 Andrew Andrianov <andrew@ncrmnt.org>, STC "Module"
https://module.ru / https://github.com/RC-MODULE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

"""

